#include "alc.h"
#include "alcboard.h"
#include "pieces.h"

#include <sys/param.h>
#include <X11/Shell.h>
#include <X11/xpm.h>
#include <X11/Xmu/Editres.h>

#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/AsciiText.h>

#if FWF
#include <Xfwf/Board.h>
#include <Xfwf/Label.h>
#include <Xfwf/Button.h>
  WidgetClass formclass = xfwfBoardWidgetClass;
  WidgetClass labelclass = xfwfLabelWidgetClass;
  WidgetClass buttonclass = xfwfButtonWidgetClass;
  String      btncallback = XtNactivate;
#elif MOTIF
#include <Xm/Xm.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h>
  WidgetClass formclass = xmRowColumnWidgetClass;
  WidgetClass labelclass = xmLabelWidgetClass;
  WidgetClass buttonclass = xmPushButtonWidgetClass;
  String      btncallback = XmNactivateCallback;
#else
  WidgetClass formclass = formWidgetClass;
  WidgetClass labelclass = labelWidgetClass;
  WidgetClass buttonclass = commandWidgetClass;
  String      btncallback = XtNcallback;
#endif 


Widget toplevel;

PLAYER players[2];
int comm;

AppData appData;
XtAppContext app_con;

#define XtNplayer1Color  "player1Color"
#define XtNplayer2Color  "player2Color"
#define XtNdefaultFile   "defaultFile"
#define XtNdebug         "debug"
#define XtNsetup         "setup"
#define XtCSound         "Sound"
#define XtNcantSound     "cantSound"
#define XtNcrunchSound   "crunchSound"
#define XtNhyperinSound  "hyperinSound"
#define XtNhyperoutSound "hyperoutSound"
#define XtNwinSound      "winSound"
#define XtNexplodeSound  "explodeSound"
#define XtNsoundCmd      "soundCmd"
#define XtNport          "port"
#define XtNhelpFile      "helpFile"

static XtResource resources[] = {
  { XtNplayer1Color, XtCColor, XtRPixel, sizeof(Pixel),
    XtOffset(AppDataPtr, player1Color),
    XtRString, (XtPointer) "red" },
  { XtNplayer2Color, XtCColor, XtRPixel, sizeof(Pixel),
    XtOffset(AppDataPtr, player2Color),
    XtRString, (XtPointer) "green" },
  { XtNdefaultFile, XtCFile, XtRString, sizeof(String),
    XtOffset(AppDataPtr, defaultFile),
    XtRString, (XtPointer) "default.xlc" },
  { XtNdebug, "Debug" , XtRBoolean, sizeof(Boolean),
    XtOffset(AppDataPtr, debug),
    XtRImmediate, (XtPointer) FALSE },
  { XtNsetup, "Setup" , XtRBoolean, sizeof(Boolean),
    XtOffset(AppDataPtr, setup),
    XtRImmediate, (XtPointer) FALSE },
  { XtNcantSound, XtCSound, XtRString, sizeof(String),
    XtOffset(AppDataPtr, cantSound),
    XtRImmediate, (XtPointer) NULL },
  { XtNcrunchSound, XtCSound, XtRString, sizeof(String),
    XtOffset(AppDataPtr, crunchSound),
    XtRImmediate, (XtPointer) NULL },
  { XtNexplodeSound, XtCSound, XtRString, sizeof(String),
    XtOffset(AppDataPtr, explodeSound),
    XtRImmediate, (XtPointer) NULL },
  { XtNhyperinSound, XtCSound, XtRString, sizeof(String),
    XtOffset(AppDataPtr, hyperinSound),
    XtRImmediate, (XtPointer) NULL },
  { XtNhyperoutSound, XtCSound, XtRString, sizeof(String),
    XtOffset(AppDataPtr, hyperoutSound),
    XtRImmediate, (XtPointer) NULL },
  { XtNwinSound, XtCSound, XtRString, sizeof(String),
    XtOffset(AppDataPtr, winSound),
    XtRImmediate, (XtPointer) NULL },
  { XtNsoundCmd, XtCString, XtRString, sizeof(String),
    XtOffset(AppDataPtr, soundCmd),
    XtRImmediate, (XtPointer) NULL},
  { XtNport, "Port", XtRInt, sizeof(int),
    XtOffset(AppDataPtr, port),
    XtRImmediate, (XtPointer) NULL},
  { XtNname, "Name", XtRString, sizeof(String),
    XtOffset(AppDataPtr, name),
    XtRImmediate, (XtPointer) NULL},
  { XtNhelpFile, XtCFile, XtRString, sizeof(String),
    XtOffset(AppDataPtr, helpFile),
    XtRImmediate, (XtPointer) "laserchess.help"},
};

XrmOptionDescRec cmd_options[] = {
  {"-setup",   ".setup",       XrmoptionNoArg, "yes" },
  {"-default", ".defaultFile", XrmoptionSepArg, NULL },
  {"-name",    ".name",        XrmoptionSepArg, NULL },
  {"-port",    ".port",        XrmoptionSepArg, NULL },
};

static String fallback_resources[] = { 
#include "fallback.h"
};

void quit()
{
  if(comm) send_quit();
  exit(0);
}

/* returns IsUnmapped = not mapped
 *         IsUnviewable = is iconified
 *         IsViewable = is mapped (but may be fully obscured)
 */
int XtGetMapState( Widget w )
{
  XWindowAttributes attr;
  XGetWindowAttributes( XtDisplay( w ), XtWindow( w ), &attr );
  return( attr.map_state );
}

#define XtIsPoppedUp XtIsMapped
Boolean XtIsMapped(Widget w)
{
  return( XtGetMapState( w ) != IsUnmapped );
}

/* tnx to J.C. Hamlin, jch@cs.umn.edu */

#define XtSetFloatArg(arg, n, d) \
    if (sizeof(float) > sizeof(XtArgVal)) { \
        XtSetArg(arg, n, &(d)); \
    } else { \
        XtArgVal *ld = (XtArgVal *)&(d); \
        XtSetArg(arg, n, *ld); \
    }

String wm_trans = "<ClientMessage>WM_PROTOCOLS: popdown()\n";
Atom wm_delete_window;
Atom wm_protocols;

void popdown_action(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    if(event->type == ClientMessage && event->xclient.type == wm_protocols
       && event->xclient.data.l[0] != wm_delete_window)
	return;
    if(w == toplevel) quit();
    else XtPopdown(w);
}

XtActionsRec actionTable[] = {
    {"popdown", popdown_action}
};

void EnableShellPopdown(Widget w)
{
  XtOverrideTranslations(w, XtParseTranslationTable(wm_trans));
  wm_protocols = XInternAtom(XtDisplay(w), "WM_PROTOCOLS", False);
  wm_delete_window = XInternAtom(XtDisplay(w), "WM_DELETE_WINDOW",
				 False);
  XSetWMProtocols(XtDisplay(w), XtWindow(w), &wm_delete_window, 1);
}

Widget helpopup,help;
Widget form,fire_b,undo_b,pass_b,save_b,load_b,quit_b,newg_b,help_b;
Widget title, subtitle;
Widget fileshell, filedialog;
int filemode;
Widget labels,moves_w[3];
AlcBoard *board;

static void cancel_cb(Widget w, XtPointer shell, XtPointer data)
{
  XtPopdown((Widget)shell);
}

static void loadsave_cb(Widget w, XtPointer shell, XtPointer data)
{
  XtPopdown(XtParent((Widget)shell));
  if(filemode == 0) board->load(XawDialogGetValueString((Widget)shell));
  else              board->save(XawDialogGetValueString((Widget)shell));
}

void CreateFileDialog(Widget label)
{
  Widget wok,wcancel;

  fileshell = XtVaCreatePopupShell("fileDialog", transientShellWidgetClass,
				 toplevel, NULL);
  filedialog = XtVaCreateManagedWidget("dialog", dialogWidgetClass, fileshell,
				       NULL);
  wcancel = 
    XtVaCreateManagedWidget("Cancel", commandWidgetClass, filedialog, NULL);
  XtAddCallback(wcancel, XtNcallback, cancel_cb, (XtPointer)fileshell );
  wok =
    XtVaCreateManagedWidget("OK", commandWidgetClass, filedialog, NULL);
  XtAddCallback(wok, XtNcallback, loadsave_cb, (XtPointer)filedialog);
  XtInstallAccelerators(filedialog, wok);
  
  XtRealizeWidget(fileshell);
  EnableShellPopdown(fileshell);
}

void PopupFileDialog(Widget label, int mode)
{
  static String s=NULL;

  if(!fileshell) CreateFileDialog(label);
  XtVaGetValues(label, XtNlabel, &s, NULL);
  XtVaSetValues(fileshell, XtNtitle, s, NULL);
  filemode = mode;
  XtPopup(fileshell, XtGrabNone);
}


static int randomtab[55];
static int randomtabi;

void goodsrandom(int seed)
{
  int i;

  randomtab[0]=seed;i=0;
  for(i++;i<55;i++){
    randomtab[i]=((randomtab[i-1]+1)*31) % 1223;
  }
  randomtabi=0;
}

int goodrandom(int range)
{
  randomtabi=(randomtabi+1)%55;
  randomtab[randomtabi]=(randomtab[(randomtabi+23)%55]+
			 randomtab[(randomtabi+54)%55]) % 1223;
  return randomtab[randomtabi] % range;
}

int playsnd(String s)
{
  char b[100];

  if(s && appData.soundCmd) {
    sprintf(b, appData.soundCmd, s);
    return system(b)>=0;
  }
  return 0;
}

void say(int msg)
{
  switch(msg){
  case SAY_CANT:
    if(!playsnd(appData.cantSound)) XBell(XtDisplay(toplevel),0);
    break;
  case SAY_CRUNCH:
    playsnd(appData.crunchSound); break;
  case SAY_EXPLODE:
    playsnd(appData.explodeSound); break;
  case SAY_HYPERIN:
    playsnd(appData.hyperinSound); break;
  case SAY_HYPEROUT:
    playsnd(appData.hyperoutSound); break;
  case SAY_WIN:
    playsnd(appData.winSound); break;
  }
}

void showmoves(AlcBoard &w)
{
  int i;

  for(i=0;i<3;i++)
    XtVaSetValues(moves_w[i],XtNbackground,
		  w.player?appData.player2Color:appData.player1Color,NULL);
  for(i=0;i<w.moves;i++) if(!XtIsManaged(moves_w[i])) XtManageChild(moves_w[i]);
  for(;i<3;i++) if(XtIsManaged(moves_w[i])) XtUnmanageChild(moves_w[i]);
}

static void quit_cb(Widget w, XtPointer object, XtPointer data)
{
  quit();
}

static void buttons_cb(Widget w, XtPointer object, XtPointer data)
{
       if(w==fire_b) board->fire_laser(1);
  else if(w==undo_b) board->undo(1);
  else if(w==pass_b) board->next_turn(1);
  else if(w==save_b) PopupFileDialog(w, 1);
  else if(w==load_b) PopupFileDialog(w, 0);
  else if(w==newg_b) board->load(appData.defaultFile);
  else if(w==help_b) {
    if(!XtIsRealized(helpopup)){
      XtRealizeWidget(helpopup);
      EnableShellPopdown(helpopup);
    }
    if(!XtIsMapped(helpopup)){
      XtPopup(helpopup,XtGrabNone);
    } else {
      XtPopdown(helpopup);
    }
  }
  else if(w==quit_b) quit();
}

void createbuttons(Widget shell)
{
  form=XtVaCreateManagedWidget("alc controls",formclass,shell,NULL);
  labels=XtVaCreateManagedWidget("labels",formclass,form,NULL);
#if FWF
  title=XtVaCreateManagedWidget("title",labelclass,form,NULL);
#endif
  subtitle=XtVaCreateManagedWidget("subtitle",labelclass,form,NULL);
  moves_w[0] = XtVaCreateManagedWidget("1",labelclass,labels,NULL);
  moves_w[1] = XtVaCreateManagedWidget("2",labelclass,labels,
			       XtNfromHoriz, moves_w[0], NULL);
  moves_w[2] = XtVaCreateManagedWidget("3",labelclass,labels,
			       XtNfromHoriz, moves_w[1], NULL);
  fire_b=XtVaCreateManagedWidget("fire",buttonclass,form,NULL);
  pass_b=XtVaCreateManagedWidget("pass",buttonclass,form,NULL);
  undo_b=XtVaCreateManagedWidget("undo",buttonclass,form,NULL);
  newg_b=XtVaCreateManagedWidget("newg",buttonclass,form,NULL);
  save_b=XtVaCreateManagedWidget("save",buttonclass,form,NULL);
  load_b=XtVaCreateManagedWidget("load",buttonclass,form,NULL);
  help_b=XtVaCreateManagedWidget("help",buttonclass,form,NULL);
  quit_b=XtVaCreateManagedWidget("quit",buttonclass,form,NULL);
}

void usage()
{
  fprintf(stderr,"Usage: laserchess [options] [opponent@computer]\n\n");
  fprintf(stderr,"options include all standard toolkit options plus:\n\n");
  fprintf(stderr,"        -setup          : run in board setup mode\n");
  fprintf(stderr,"        -default [file] : use [file] as default board\n");
  fprintf(stderr,"                        : default is %s\n",appData.defaultFile);
  fprintf(stderr,"        -name [name]    : use [name] as your name\n"); 
  fprintf(stderr,"                        : default is your login name\n");
  fprintf(stderr,"        -port [port]    : use [port] as comm port\n");
  fprintf(stderr,"                        : default is hash value calculated\n");
  fprintf(stderr,"                        : from your name\n");
}

main(int argc,char **argv)
{
  Widget box,canvas;
  int pid;
  char s[100];

  pid = getpid();

  toplevel = XtAppInitialize (&app_con, "Laserchess",
			      cmd_options, XtNumber(cmd_options),
			      &argc, argv,
			      fallback_resources, NULL,ZERO);

  XtGetApplicationResources(toplevel,&appData,resources,XtNumber(resources),
			    NULL,0);
  
  if (argc >= 2 && argv[1][0]=='-') {
    usage();
    exit(1);
  }
  char *at_sym;
  if (argc == 2) {
    if(at_sym = strchr(argv[1], '@')) *at_sym++ = '\0';
    setup_communication (appData.name, argv[1], at_sym);
    EnemyID = XtAppAddInput (app_con, players[1].fd, (XtPointer)XtInputReadMask,
			     EnemyInterrupt, NULL);
    comm = 1;
    if(players[0].has_first_move){
      goodsrandom(pid);
      send_srand(pid);
    }
    sprintf(s,"%s vs. %s", players[0].name, players[1].name);
    printf("%s\n",players[0].has_first_move?"HasFirstMove":"HasNOTfirstMove");
  } else if(argc != 1) {
    usage();
    exit(1);
  } else {
    goodsrandom(pid);
    strcpy(s,"");
  }


#if MOTIF
  XtAddEventHandler(toplevel, (EventMask)0, True,
		    (XtEventHandler)_XEditResCheckMessages, 0);
#endif

  XtAppAddActions(app_con, actionTable, XtNumber(actionTable));

  box=XtCreateManagedWidget("form",panedWidgetClass,toplevel,NULL,ZERO);
  createbuttons(box);
  canvas=XtVaCreateManagedWidget("alcBoard",alcBoardWidgetClass, box,
				 XtNboardWidth, COLS,
				 XtNboardHeight, ROWS,
				 NULL);

  helpopup = XtCreatePopupShell("helpShell", transientShellWidgetClass,
				toplevel, NULL, 0);

  help = XtVaCreateManagedWidget("help", asciiTextWidgetClass, helpopup,
				 XtNeditType, XawtextRead,
				 NULL);

  XtVaSetValues(help, XtNtype, XawAsciiFile, XtNstring, appData.helpFile);

  AlcBoard board(canvas,showmoves);
  ::board = &board;

  XtAddCallback(fire_b,btncallback,buttons_cb,&board);
  XtAddCallback(undo_b,btncallback,buttons_cb,&board);
  XtAddCallback(pass_b,btncallback,buttons_cb,&board);
  XtAddCallback(save_b,btncallback,buttons_cb,&board);
  XtAddCallback(load_b,btncallback,buttons_cb,&board);
  XtAddCallback(newg_b,btncallback,buttons_cb,&board);
  XtAddCallback(help_b,btncallback,buttons_cb,&board);
  XtAddCallback(quit_b,btncallback,buttons_cb,&board);
  XtVaSetValues(subtitle, XtNlabel, s, NULL);

  XtRealizeWidget(toplevel);

  EnableShellPopdown(toplevel);

  board.setup=appData.setup;
  board.load(appData.defaultFile);

  XtAppMainLoop(app_con);
}
