/* DrawingA.c: The AlcBoard Widget Methods */

/* Copyright 1990, David Nedde   (XawDrawingAreaWidget on which this is based)
/* Copyright 1994, Ziga Kranjec
/*
/* Permission to use, copy, modify, and distribute this
/* software and its documentation for any purpose and without fee
/* is granted provided that the above copyright notice appears in all copies.
/* It is provided "as is" without express or implied warranty.
*/

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#include <X11/Xaw/SimpleP.h>
#include "alcboardP.h"

static void	Initialize();
static void	Destroy();
static void     Realize();
static void     DrawSquare();
static Boolean  SetValues();
static void     AdjustSize();
static void     Redraw();
static void     Resize();
static void	Redisplay();
static void	input_draw();
static void	motion_draw();
static void	resize_draw();
static void     select_draw();
static void     adjust_draw();
static void     set_hole_background();

static char defaultTranslations[] = "<Btn1Down>: select() \n <Btn2Down>: adjust() \n <KeyDown>: input() \n <KeyUp>: input() \n <Motion>: motion() \n <Configure>: resize()";
static XtActionsRec actionsList[] = {
  { "select", (XtActionProc)select_draw },
  { "adjust", (XtActionProc)adjust_draw },
  { "input",  (XtActionProc)input_draw },
  { "motion", (XtActionProc)motion_draw },
  { "resize", (XtActionProc)resize_draw },
};

/* Default instance record values */
static XtResource resources[] = {
  {XtNboardWidth, XtCWidth, XtRDimension, sizeof(Dimension),
     XtOffset(AlcBoardWidget, board.boardWidth),
     XtRImmediate, (XtPointer)15 },
  {XtNboardHeight, XtCHeight, XtRDimension, sizeof(Dimension),
     XtOffset(AlcBoardWidget, board.boardHeight),
     XtRImmediate, (XtPointer)11 },
  {XtNsquareSize, XtCLength, XtRDimension, sizeof(Dimension),
     XtOffset(AlcBoardWidget, board.squareSize),
     XtRImmediate, (XtPointer)55 },
  {XtNsquareBorderWidth, XtCBorderWidth, XtRDimension, sizeof(Dimension),
     XtOffset(AlcBoardWidget, board.squareBorderWidth),
     XtRImmediate, (XtPointer)1 },
  {XtNsquareBorderColor, XtCBorderColor, XtRPixel, sizeof(Pixel),
     XtOffset(AlcBoardWidget, board.squareBorderColor),
     XtRString, (String)"black" },
  {XtNdarkBackground, XtCBackground, XtRPixel, sizeof(Pixel),
     XtOffset(AlcBoardWidget, board.darkBackground),
     XtRString, (String)"gray33" },
  {XtNlightBackground, XtCBackground, XtRPixel, sizeof(Pixel),
     XtOffset(AlcBoardWidget, board.lightBackground),
     XtRString, (String)"gray66" },
  {XtNselectBackground, XtCBackground, XtRPixel, sizeof(Pixel),
     XtOffset(AlcBoardWidget, board.selectBackground),
     XtRString, (String)"deep pink" },
  {XtNholeBackground, XtCBackground, XtRPixel, sizeof(Pixel),
     XtOffset(AlcBoardWidget, board.holeBackground),
     XtRCallProc, (XtPointer) set_hole_background },
  {XtNfrozenColor, XtCForeground, XtRPixel, sizeof(Pixel),
     XtOffset(AlcBoardWidget, board.frozenColor),
     XtRString, (String)"snow" },
  {XtNlaserColor, XtCForeground, XtRPixel, sizeof(Pixel),
     XtOffset(AlcBoardWidget, board.laserColor),
     XtRString, (String)"yellow" },
  {XtNlaserThickness, XtCThickness, XtRDimension, sizeof(Dimension),
     XtOffset(AlcBoardWidget, board.laserThickness),
     XtRImmediate, (XtPointer)1 },
  {XtNiconFile, XtCFile, XtRString, sizeof(String),
     XtOffset(AlcBoardWidget, board.iconFile),
     XtRImmediate, (XtPointer)0 },
  {XtNselectCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
     XtOffset(AlcBoardWidget, board.select_callback), 
     XtRCallback, NULL },
  {XtNadjustCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
     XtOffset(AlcBoardWidget, board.adjust_callback), 
     XtRCallback, NULL },
  {XtNexposeCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
     XtOffset(AlcBoardWidget, board.expose_callback), 
     XtRCallback, NULL },
  {XtNinputCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
     XtOffset(AlcBoardWidget, board.input_callback), 
     XtRCallback, NULL },
  {XtNmotionCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
     XtOffset(AlcBoardWidget, board.motion_callback), 
     XtRCallback, NULL },
  {XtNresizeCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
     XtOffset(AlcBoardWidget, board.resize_callback), 
     XtRCallback, NULL },
};


AlcBoardClassRec alcBoardClassRec = {
  /* CoreClassPart */
{
  (WidgetClass) &simpleClassRec,	/* superclass		  */	
    "AlcBoard", 			/* class_name		  */
    sizeof(AlcBoardRec),		/* size			  */
    NULL,				/* class_initialize	  */
    NULL,				/* class_part_initialize  */
    FALSE,				/* class_inited		  */
    Initialize,				/* initialize		  */
    NULL,				/* initialize_hook	  */
    Realize,		        	/* realize		  */
    actionsList,			/* actions		  */
    XtNumber(actionsList),		/* num_actions		  */
    resources,				/* resources		  */
    XtNumber(resources),		/* resource_count	  */
    NULLQUARK,				/* xrm_class		  */
    FALSE,				/* compress_motion	  */
    TRUE,				/* compress_exposure	  */
    TRUE,				/* compress_enterleave    */
    FALSE,				/* visible_interest	  */
    Destroy,				/* destroy		  */
    Resize,				/* resize		  */
    Redisplay,				/* expose		  */
    SetValues,				/* set_values		  */
    NULL,				/* set_values_hook	  */
    XtInheritSetValuesAlmost,		/* set_values_almost	  */
    NULL,				/* get_values_hook	  */
    NULL,				/* accept_focus		  */
    XtVersion,				/* version		  */
    NULL,				/* callback_private	  */
    defaultTranslations,		/* tm_table		  */
    XtInheritQueryGeometry,		/* query_geometry	  */
    XtInheritDisplayAccelerator,	/* display_accelerator	  */
    NULL				/* extension		  */
  },  /* CoreClass fields initialization */
  {
    /* change_sensitive		*/	XtInheritChangeSensitive
  },  /* SimpleClass fields initialization */
  {
    0,                                     /* field not used    */
  },  /* AlcBoardClass fields initialization */
};

  
WidgetClass alcBoardWidgetClass = (WidgetClass)&alcBoardClassRec;

static void Initialize( request, new)
AlcBoardWidget request, new;
{
  Dimension bw;

  bw=request->board.squareBorderWidth;

  if (request->core.width == 0)
    new->core.width = request->board.boardWidth*(bw+request->board.squareSize) + bw;
  if (request->core.height == 0)
    new->core.height = request->board.boardHeight*(bw+request->board.squareSize) + bw;

  Resize(new);

  new->board.icons=NULL;
  new->board.square = (int *)XtCalloc(sizeof(*new->board.square),
			      request->board.boardWidth *
			      request->board.boardHeight);
  new->board.selx = new->board.sely = -1;
}


static void Destroy( w)
AlcBoardWidget w;
{
  XtFree((void *)w->board.square);
/*
  XtReleaseGC((Widget)w, w->board.lasergc);
  XtReleaseGC((Widget)w, w->board.blitgc);
  XtReleaseGC((Widget)w, w->board.squarebordergc);
  XtReleaseGC((Widget)w, w->board.lightgc);
  XtReleaseGC((Widget)w, w->board.darkgc);
*/
  XtRemoveAllCallbacks((Widget)w, XtNexposeCallback);
  XtRemoveAllCallbacks((Widget)w, XtNinputCallback);
  XtRemoveAllCallbacks((Widget)w, XtNmotionCallback);
  XtRemoveAllCallbacks((Widget)w, XtNresizeCallback);
}

/* Realize */
static void Realize(w, valueMask, attributes)
AlcBoardWidget w;
XtValueMask *valueMask;
XSetWindowAttributes *attributes;
{
  XGCValues values;

  alcBoardClassRec.core_class.superclass->core_class.realize
    ((Widget)w, valueMask, attributes);

  values.foreground=w->board.lightBackground;
  w->board.lightgc = XCreateGC(XtDisplay(w), XtWindow(w), 
				 (GCForeground), &values);
  values.foreground=w->board.darkBackground;
  w->board.darkgc = XCreateGC(XtDisplay(w), XtWindow(w), 
			      (GCForeground), &values);
  values.foreground=w->board.squareBorderColor;
  values.line_width=w->board.squareBorderWidth;
  w->board.squarebordergc = XCreateGC(XtDisplay(w), XtWindow(w), 
					(GCForeground|GCLineWidth),
					&values);
  values.function=GXcopy;
  values.foreground=w->board.frozenColor;
  w->board.blitgc = XCreateGC(XtDisplay(w), XtWindow(w), 
				(GCFunction), &values);

  values.foreground=w->board.laserColor;
  values.line_width=w->board.laserThickness;
  values.join_style=JoinBevel;
  w->board.lasergc = XCreateGC(XtDisplay(w), XtWindow(w), 
				 (GCForeground|GCLineWidth|GCJoinStyle),
				 &values);

  AlcBoardLoadIcons(w,w->board.iconFile);
}

static void set_hole_background(w,o,v)
AlcBoardWidget w;
int o;
XrmValue *v;
{
  v->addr = (XtPointer)(&w->core.background_pixel);
}

/**** Set Values ****/
static Boolean SetValues(current, request, new)
AlcBoardWidget current, request, new;
{
  int Redraw,XtRedraw;
  Redraw=XtRedraw=0;

#define ifchanged(a) if(current->a != new->a)

  ifchanged(board.boardWidth){
    XtWarning("AlcBoardWidget: can't change boardWidth with XtSetValues.\n");
    new->board.boardWidth = current->board.boardWidth;
  }
  ifchanged(board.boardHeight){
    XtWarning("AlcBoardWidget: can't change boardHeight with XtSetValues.\n");
    new->board.boardHeight = current->board.boardHeight;
  }
  ifchanged(board.squareSize){}
  ifchanged(board.squareBorderWidth){}
  ifchanged(board.squareBorderColor){
    XSetForeground(XtDisplay(request),
		    new->board.squarebordergc,
		    new->board.squareBorderColor);
    XtRedraw=1;
  }
  ifchanged(board.darkBackground){
    XSetForeground(XtDisplay(request),
		    new->board.darkgc,
		    new->board.darkBackground);
    Redraw=1;
  }
  ifchanged(board.lightBackground){
    XSetForeground(XtDisplay(request),
		    new->board.lightgc,
		    new->board.lightBackground);
    Redraw=1;
  }
  ifchanged(board.holeBackground){ Redraw=1; }
  ifchanged(board.frozenColor){
    XSetForeground(XtDisplay(request),
		   new->board.blitgc,
		   new->board.frozenColor);
    Redraw=1;
  }
  ifchanged(board.laserColor){
    XSetForeground(XtDisplay(request),
		   new->board.lasergc,
		   new->board.laserColor);
  }
  ifchanged(board.laserThickness){
    XSetLineAttributes(XtDisplay(request),
		       new->board.lasergc,
		       new->board.laserThickness,
		       LineSolid, CapButt, JoinBevel);
  }
  ifchanged(board.iconFile){
    AlcBoardLoadIcons(new,new->board.iconFile);
    XtRedraw=1;
  }
  if(Redraw) AlcBoardRefresh(new);
  return(XtRedraw && !Redraw);

#undef ifchanged(a)
}

/**** Sizing *****/

/* AdjustSize */
static void AdjustSize(w, width_in, height_in, width_out, height_out, sqsize_out)
AlcBoardWidget w;
Dimension width_in,height_in;
Dimension *width_out,*height_out,*sqsize_out;
{
  int bw,sqs1,sqs2,sqs;

  bw=w->board.squareBorderWidth;
  sqs1=(width_in-bw) / w->board.boardWidth - bw;
  sqs2=(height_in-bw) / w->board.boardHeight - bw;
  sqs=sqs1<sqs2?sqs1:sqs2; if(!(sqs&1)) sqs--;
  if(width_out) *width_out = bw + (bw+sqs)*w->board.boardWidth;
  if(height_out) *height_out = bw + (bw+sqs)*w->board.boardHeight;
  if(sqsize_out) *sqsize_out = sqs;
}

/* resize method */
static void Resize(w)
AlcBoardWidget w;
{
  Dimension W,H;

  AdjustSize(w, w->core.width, w->core.height, &W, &H, &w->board.squareSize);
  w->board.x0=(w->core.width-W)/2+w->board.squareBorderWidth/2;
  w->board.y0=(w->core.height-H)/2+w->board.squareBorderWidth/2;
}

/***** Drawing *****/

/* redraw single square */
static void DrawSquare(w, x, y, r, what)
AlcBoardWidget w;
int x,y;
Region r;
int what;
{
  int W,i,ii,x0,y0,x1,y1;
  int w1,h1;

  if( x >= w->board.boardWidth || y >= w->board.boardHeight) return;

  W=w->board.squareBorderWidth+w->board.squareSize;
  x0 = w->board.x0+x*W;
  y0 = w->board.y0+y*W;

  if( r && XRectInRegion(r,x0,y0,W,W)==RectangleOut ) return;

  if(what & 2){
    if(w->board.square && w->board.icons){
      i=(w->board.boardHeight)*x+y;
      ii=w->board.square[i]&0x200;
    } else ii=0;
    if(x==w->board.selx && y==w->board.sely){
      XSetForeground(XtDisplay(w), w->board.darkgc, w->board.selectBackground);
      XFillRectangle(XtDisplay(w), XtWindow(w),w->board.darkgc, x0, y0, W, W);
      XSetForeground(XtDisplay(w), w->board.darkgc, w->board.darkBackground);
    } else if(ii) {
      XSetForeground(XtDisplay(w), w->board.darkgc, w->board.holeBackground);
      XFillRectangle(XtDisplay(w), XtWindow(w),w->board.darkgc, x0, y0, W, W);
      XSetForeground(XtDisplay(w), w->board.darkgc, w->board.darkBackground);
    } else {
      XFillRectangle(XtDisplay(w), XtWindow(w),
		     ((x+y)&1)?w->board.lightgc:w->board.darkgc, x0, y0, W, W);
    }
    XDrawRectangle(XtDisplay(w), XtWindow(w), w->board.squarebordergc,
		   x0, y0, W, W);
    if(ii) XDrawRectangle(XtDisplay(w), XtWindow(w), w->board.squarebordergc,
			  x0, y0 + w->board.squareBorderWidth,
			  W - w->board.squareBorderWidth,
			  W - w->board.squareBorderWidth );
  }

  if(what & 1){
    if(w->board.square && w->board.icons){
      i=(w->board.boardHeight)*x+y;
      ii=w->board.square[i]&0x100;
      i=w->board.square[i]&0xff;
      if( i > 0 && i < w->board.nicons){

	i -= 1;
	w1 = w->board.iconPixmapAttr.x_hotspot;
	h1 = w->board.iconPixmapAttr.y_hotspot;
	x1 = x0 + (W - w1 + 1)/2;
	y1 = y0 + (W - h1 + 1)/2;
	
	XSetClipMask(XtDisplay(w), w->board.blitgc, w->board.icons[i].mask);
	XSetClipOrigin(XtDisplay(w), w->board.blitgc, x1, y1);
	
	if(ii){
	  XFillRectangle(XtDisplay(w), XtWindow(w),w->board.blitgc, x0, y0, W, W);
	} else {
	  XCopyArea(XtDisplay(w), w->board.icons[i].pix, XtWindow(w), 
		    w->board.blitgc, 0, 0, w1, h1, x1, y1);
	}
      }
    }
  }
}

static void Refresh(w, r)
AlcBoardWidget w;
Region r;
{
  int i,j;

  if(!XtIsRealized((Widget)w)) return;


  for(j=0;j<w->board.boardHeight;j++)
    for(i=0;i<w->board.boardWidth;i++) DrawSquare(w,i,j,r,3);
}

void AlcBoardRefresh(w)
AlcBoardWidget w;
{  Refresh(w, NULL); }

void AlcBoardRefreshSquare(w, x, y, mode)
AlcBoardWidget w;
int x,y;
int mode;
{  DrawSquare(w, x, y, NULL, mode?1:3); }

/* redraw & Invoke expose callbacks */
static void Redisplay(w, event, region)
AlcBoardWidget    w;
XEvent		  *event;
Region		  region;
{
  AlcBoardCallbackStruct cb;

  Refresh(w, region);

  cb.reason = alcbCR_EXPOSE;
  cb.event  = event;
  cb.window = XtWindow(w);
  XtCallCallbacks((Widget)w, XtNexposeCallback, (char *)&cb);
}

/* convert pixel to square coordinates */
static int xy2xcyc(w,x,y,xc,yc)
AlcBoardWidget w;
int x,y,*xc,*yc;
{
  x -= w->board.x0;
  y -= w->board.y0;
  *xc=x/(w->board.squareSize+w->board.squareBorderWidth);
  *yc=y/(w->board.squareSize+w->board.squareBorderWidth);
  if(*xc>=0 && *xc<w->board.boardWidth &&
     *yc>=0 && *yc<w->board.boardHeight && x>=0 && y>=0)
    return 1;
  else
    return 0;
}

/* select action */
static void select_draw(w, event, args, n_args)
AlcBoardWidget w;
XEvent *event;
char *args[];
int n_args;
{
  int x,y;
  AlcBoardCallbackStruct cb;
  if(event->type == ButtonPress || event->type == ButtonRelease){
    x=((XButtonEvent *)event)->x;
    y=((XButtonEvent *)event)->y;
    if(xy2xcyc(w,x,y,&cb.x,&cb.y)){
      cb.reason = alcbCR_SELECT;
      cb.event  = event;
      cb.window = XtWindow(w);
      XtCallCallbacks((Widget)w, XtNselectCallback, (char *)&cb);
    }
  } else {
    XtWarningMsg(NULL,"bad event for select() action",
		 XtCEventBindings,"AlcBoardWidget::select():BadEvent",
		 NULL, 0);
  }
}

/* adjust action */
static void adjust_draw(w, event, args, n_args)
AlcBoardWidget w;
XEvent *event;
char *args[];
int n_args;
{
  int x,y;
  AlcBoardCallbackStruct cb;
  if(event->type == ButtonPress || event->type == ButtonRelease){
    x=((XButtonEvent *)event)->x;
    y=((XButtonEvent *)event)->y;
    xy2xcyc(w,x,y,&cb.x,&cb.y);
    cb.reason = alcbCR_ADJUST;
    cb.event  = event;
    cb.window = XtWindow(w);
    XtCallCallbacks((Widget)w, XtNadjustCallback, (char *)&cb);
  } else {
    XtWarningMsg(NULL,"bad event for adjust() action",
		 XtCEventBindings,"AlcBoardWidget::adjust():BadEvent",
		 NULL, 0);
  }
}

/* Invoke resize callbacks */
static void resize_draw(w, event, args, n_args)
AlcBoardWidget w;
XEvent		 *event;
char		 *args[];
int		  n_args;
{
  AlcBoardCallbackStruct cb;

  cb.reason = alcbCR_RESIZE;
  cb.event  = event;
  cb.window = XtWindow(w);
  XtCallCallbacks((Widget)w, XtNresizeCallback, (char *)&cb);
}

/* Invoke input callbacks */
static void input_draw(w, event, args, n_args)
AlcBoardWidget w;
XEvent		 *event;
char		 *args[];
int		  n_args;
{
  AlcBoardCallbackStruct cb;

  cb.reason = alcbCR_INPUT;
  cb.event  = event;
  cb.window = XtWindow(w);
  XtCallCallbacks((Widget)w, XtNinputCallback, (char *)&cb);
}

/* Invoke motion callbacks */
static void motion_draw(w, event, args, n_args)
AlcBoardWidget w;
XEvent		 *event;
char		 *args[];
int		  n_args;
{
  AlcBoardCallbackStruct cb;

  cb.reason = alcbCR_MOTION;
  cb.event  = event;
  cb.window = XtWindow(w);
  XtCallCallbacks((Widget)w, XtNmotionCallback, (char *)&cb);
}

int AlcBoardLoadIcons(w, filename)
AlcBoardWidget w;
String filename;
{
  Pixmap pix,mask;
  XpmAttributes attr;
  XpmExtension *p;
  int ii,i,j,err;
  int nx,ny,w1,h1,x1,y1;
  GC gc,gc1;

  if(!filename || filename[0]=='\0') return 0;

  attr.closeness = 65536/32;
  attr.valuemask = XpmCloseness;
  err = XpmReadFileToPixmap(XtDisplay(w),XtWindow(w),filename,&pix,&mask,&attr);
  if(err) switch(err) {
  case XpmOpenFailed:
    XtWarningMsg(filename, "cannot open file",
		 XtCFile, "AlcBoardLoadIcons:XpmOpenFailed",
		 NULL,0);
    return 0;
  case XpmFileInvalid:
    XtWarningMsg(filename, "bad xpm file",
		 XtCFile, "AlcBoardLoadIcons:XpmFileInvalid",
		 NULL,0);
    return 0;
  case XpmNoMemory:
    XtWarningMsg(filename, "brain drain",
		 XtCFile, "AlcBoardLoadIcons:XpmNoMemory",
		 NULL,0);
    return 0;
  case XpmColorError:
    XtWarningMsg(filename, "cannot allocate colors",
		 XtCFile, "AlcBoardLoadIcons:XpmColorError",
		 NULL, 0);
    return 0;
  }

  if(!mask) {
    XtWarningMsg(filename, "xpm file needs a mask (color None)",
		 XtCFile, "AlcBoardLoadIcons:NoMask",
		 NULL,0);
    goto FAIL_FREE;
  }

  if(!(attr.valuemask&XpmHotspot)|| attr.x_hotspot == 0 || attr.y_hotspot == 0) {
    XtWarningMsg(filename, "xpm file needs hotspot greater than (0,0)",
		 XtCFile, "AlcBoardLoadIcons:BadHotspot",
		 NULL,0);
    XFreePixmap(XtDisplay(w),mask);
    goto FAIL_FREE;
  }
    
  if(w->board.icons){
    for(i=0;i<w->board.nicons;i++){
      XFreePixmap(XtDisplay(w),w->board.icons[i].pix);
      XFreePixmap(XtDisplay(w),w->board.icons[i].mask);
    }
    XtFree((void *)w->board.icons);
    XpmFreeAttributes(&w->board.iconPixmapAttr);
  }

  w->board.iconPixmapAttr=attr;
  nx = attr.width/attr.x_hotspot;
  ny = attr.height/attr.y_hotspot;
  w->board.nicons=nx*ny;
  w->board.icons=(AlcBoardIcon *)XtCalloc(nx*ny,sizeof(AlcBoardIcon));

  w1 = w->board.iconPixmapAttr.x_hotspot;
  h1 = w->board.iconPixmapAttr.y_hotspot;

  gc  = XCreateGC(XtDisplay(w), XtWindow(w), 0, NULL);
  gc1 = XCreateGC(XtDisplay(w), mask, 0, NULL);

  for(ii=j=0;j<ny;j++) for(i=0;i<nx;i++,ii++){
    x1 = i * w1;
    y1 = j * h1;

    w->board.icons[ii].pix=XCreatePixmap(XtDisplay(w),XtWindow(w),
					 w1, h1, w->core.depth);
    w->board.icons[ii].mask=XCreatePixmap(XtDisplay(w),XtWindow(w),
					 w1, h1, 1);
    XCopyArea(XtDisplay(w), pix, w->board.icons[ii].pix, 
	      gc, x1, y1, w1, h1, 0, 0);
    XCopyArea(XtDisplay(w), mask, w->board.icons[ii].mask,
	      gc1, x1, y1, w1, h1, 0, 0);
  }

  XFreeGC(XtDisplay(w), gc1);
  XFreeGC(XtDisplay(w), gc);
  XFreePixmap(XtDisplay(w), pix);
  XFreePixmap(XtDisplay(w), mask);
  
  return 1;

 FAIL_FREE:
  XFreePixmap(XtDisplay(w),pix);
  XpmFreeAttributes(&attr);
  return 0;
}

void AlcBoardSetSquare(w, x, y, iconNum)
AlcBoardWidget w;
int x,y,iconNum;
{
  int i;

  if( x >= w->board.boardWidth || y >= w->board.boardHeight) return;
  i=(w->board.boardHeight)*x+y;
  w->board.square[i]=iconNum;
}

void AlcBoardSelectSquare(w, x, y, select)
AlcBoardWidget w;
int x,y,select;
{
  int x0,y0;

  x0=w->board.selx;
  y0=w->board.sely;
  if(!select) {
    w->board.selx=w->board.sely=-1;
  } else {
    w->board.selx=x;
    w->board.sely=y;
    AlcBoardRefreshSquare(w,x,y,0);
  }
  if(x0!=-1) AlcBoardRefreshSquare(w,x0,y0,0);
}

void AlcBoardDrawLaserSegments(W, x, y, segments)
AlcBoardWidget W;
int x,y,segments;
{
  int w,w2,x0,y0;
  w = W->board.squareSize+W->board.squareBorderWidth;
  w2 = w/2;
  x0=W->board.x0+x*w;
  y0=W->board.y0+y*w;
  if(segments & alcbSEG_N ) XDrawLine(XtDisplay(W), XtWindow(W), W->board.lasergc,
				      x0+w2, y0+w2, x0+w2, y0);
  if(segments & alcbSEG_NE) XDrawLine(XtDisplay(W), XtWindow(W), W->board.lasergc,
				      x0+w2, y0+w2, x0+w, y0);
  if(segments & alcbSEG_E ) XDrawLine(XtDisplay(W), XtWindow(W), W->board.lasergc,
				      x0+w2, y0+w2, x0+w, y0+w2);
  if(segments & alcbSEG_SE) XDrawLine(XtDisplay(W), XtWindow(W), W->board.lasergc,
				      x0+w2, y0+w2, x0+w, y0+w);
  if(segments & alcbSEG_S ) XDrawLine(XtDisplay(W), XtWindow(W), W->board.lasergc,
				      x0+w2, y0+w2, x0+w2, y0+w);
  if(segments & alcbSEG_SW) XDrawLine(XtDisplay(W), XtWindow(W), W->board.lasergc,
				      x0+w2, y0+w2, x0, y0+w);
  if(segments & alcbSEG_W ) XDrawLine(XtDisplay(W), XtWindow(W), W->board.lasergc,
				      x0+w2, y0+w2, x0, y0+w2);
  if(segments & alcbSEG_NW) XDrawLine(XtDisplay(W), XtWindow(W), W->board.lasergc,
				      x0+w2, y0+w2, x0, y0);
}

static int AlcBoardRandom(int range)
{
  int r;
  r=rand()%range;
  return r;
}

/* draw explosions in widget w,
   on n squares stored in array p */
void AlcBoardFireworks(w, n, p)
AlcBoardWidget w;
int n;
int (*p)[2];
{
  typedef struct {
    int vx,vy,age;
  } pointdata;

  XPoint *a,*b,*t;
  pointdata *pd;
  int x0,y0,x1,y1,i,j,k,W;
  GC gc;
  XGCValues values;

  W=w->board.squareBorderWidth+w->board.squareSize;
  a=(XPoint *)XtCalloc(n*ALC_FIREWORKS_POINTS,sizeof(XPoint));
  b=(XPoint *)XtCalloc(n*ALC_FIREWORKS_POINTS,sizeof(XPoint));
  pd=(pointdata *)XtCalloc(n*ALC_FIREWORKS_POINTS,sizeof(pointdata));
  values.function = GXxor;
  values.foreground = 0;
  gc = XtGetGC((Widget)w,(GCForeground),&values);

  for(k=i=0;i<n;i++){
    x0=w->board.x0+p[i][0]*W;
    y0=w->board.y0+p[i][1]*W;
    for(j=0;j<ALC_FIREWORKS_POINTS;j++){
      a[k].x=x0+AlcBoardRandom(W);
      a[k].y=y0+AlcBoardRandom(W);
      pd[k].vx=(a[k].x-(x0+W/2))/(W/9)+AlcBoardRandom(5)-2;
      pd[k].vy=(a[k].y-(y0+W/2))/(W/9)+AlcBoardRandom(5)-2;
      pd[k].age=1+AlcBoardRandom(ALC_FIREWORKS_STEPS);
      k++;
    }
  }
  x0 = w->board.x0;
  y0 = w->board.y0;
  x1 = x0+W*w->board.boardWidth;
  y1 = y0+W*w->board.boardHeight;
  n *= ALC_FIREWORKS_POINTS;

  XDrawPoints(XtDisplay(w),XtWindow(w),gc,a,n,CoordModeOrigin);
  while(n>0){
    for(i=j=0;i<n;i++){
      if((--pd[i].age)<=0) continue;
      b[j].x=a[i].x-pd[i].vx;
           if(b[j].x<x0) { b[j].x = 2*x0-b[j].x; pd[i].vx = -pd[i].vx; }
      else if(b[j].x>x1) { b[j].x = 2*x1-b[j].x; pd[i].vx = -pd[i].vx; }
      b[j].y=a[i].y-pd[i].vy;
           if(b[j].y<y0) { b[j].y = 2*y0-b[j].y; pd[i].vy = -pd[i].vy; }
      else if(b[j].y>y1) { b[j].y = 2*y1-b[j].y; pd[i].vy = -pd[i].vy; }
      pd[j]=pd[i]; j++;
    }
    XDrawPoints(XtDisplay(w),XtWindow(w),gc,a,n,CoordModeOrigin);
    n=j;
    XDrawPoints(XtDisplay(w),XtWindow(w),gc,b,n,CoordModeOrigin);
    XFlush(XtDisplay(w));
    usleep(100);
    t=a;a=b;b=t;
  }

  XtReleaseGC((Widget)w,gc);
  XtFree((void *)pd);XtFree((void *)b);XtFree((void *)a);
  AlcBoardRefresh(w);
}














