/* alcboard.h - Public Header file */

/* Copyright 1990, David Nedde   (XawDrawingAreaWidget on which this is based)
/* Copyright 1994, Ziga Kranjec
/*
/* Permission to use, copy, modify, and distribute this
/* software and its documentation for any purpose and without fee
/* is granted provided that the above copyright notice appears in all copies.
/* It is provided "as is" without express or implied warranty.
*/

/* Define widget's class pointer and strings used to specify resources */

#ifndef _AlcBoard_h
#define _AlcBoard_h

#ifdef __cplusplus
#define LINKAGE "C"
#else
#define LINKAGE
#endif

/* Resources ADDED to label widget:

 Name		     Class		RepType		Default Value
 ----		     -----		-------		-------------
 boardWidth          Width              Dimension       15
 boardHeight         Height             Dimension       11
 squareSize          Length             Dimension       55
 squareBorderWidth   BorderWidth	Dimension	1
 squareBorderColor   BorderColor        Pixel           black
 darkBackground      Background         Pixel           gray33
 lightBackground     Background         Pixel           gray66
 selectBackground    Background         Pixel           deep pink
 holeBackground      Background         Pixel           core->background_pixel
 frozenColor         Foregorund         Pixel           snow
 laserColor          Foregorund         Pixel           yellow
 laserThickness      Thickness          Dimension       3
 iconFile            File               String          NULL
 selectCallback      Callback           Pointer         NULL
 adjustCallback      Callback           Pointer         NULL
 exposeCallback	     Callback		Pointer		NULL
 inputCallback	     Callback		Pointer		NULL
 motionCallback	     Callback		Pointer		NULL
 resizeCallback	     Callback		Pointer		NULL
*/


extern WidgetClass alcBoardWidgetClass;

typedef struct _AlcBoardClassRec *AlcBoardWidgetClass;
typedef struct _AlcBoardRec	    *AlcBoardWidget;


/* Resource strings */
#define XtNboardWidth           "boardWidth"
#define XtNboardHeight          "boardHeight"
#define XtNsquareSize           "squareSize"
#define XtNsquareBorderWidth    "squareBorderWidth"
#define XtNsquareBorderColor    "squareBorderColor"
#define XtNdarkBackground       "darkBackground"
#define XtNlightBackground      "lightBackground"
#define XtNselectBackground     "selectBackground"
#define XtNholeBackground       "holeBackground"
#define XtNfrozenColor          "frozenColor"
#define XtNlaserColor           "laserColor"
#define XtNlaserThickness       "laserThickness"
#define XtNiconFile             "iconFile"
#define XtNselectCallback       "selectCallback"
#define XtNadjustCallback       "adjustCallback"
#define XtNexposeCallback	"exposeCallback"
#define XtNinputCallback	"inputCallback"
#define XtNmotionCallback	"motionCallback"
#define XtNresizeCallback	"resizeCallback"


typedef struct _AlcBoardCallbackStruct {
  int	  reason;
  XEvent *event;
  Window  window;
  int     x,y;       /* square */
} AlcBoardCallbackStruct;

/* Reasons */
#define alcbCR_EXPOSE 1
#define alcbCR_INPUT  2
#define alcbCR_MOTION 3
#define alcbCR_RESIZE 4
#define alcbCR_SELECT 5
#define alcbCR_ADJUST 6

/* laser segments */
#define alcbSEG_N  (1L<<0)
#define alcbSEG_NE (1L<<1)
#define alcbSEG_E  (1L<<2)
#define alcbSEG_SE (1L<<3)
#define alcbSEG_S  (1L<<4)
#define alcbSEG_SW (1L<<5)
#define alcbSEG_W  (1L<<6)
#define alcbSEG_NW (1L<<7)

/* utility functions */

extern LINKAGE void AlcBoardRefresh();
extern LINKAGE void AlcBoardRefreshSquare();
extern LINKAGE int  AlcBoardLoadIcons();
extern LINKAGE void AlcBoardSetSquare();
extern LINKAGE void AlcBoardSelectSquare();
extern LINKAGE void AlcBoardDrawLaserSegments();
extern LINKAGE void AlcBoardFireworks();

#undef LINKAGE
#endif /* _AlcBoard_h */


