/*
 * Message.c
 *
 * Player to player messages module.
 */

#include "alc.h"
#include "pieces.h"

#include <stdarg.h>

XtInputId EnemyID;

#define T_SELECT	'0'		/* Messages during play */
#define T_ROTATE	'1'
#define T_FIRE          '2'
#define T_NEXT_TURN     '3'
#define T_SRAND         '4'
#define T_UNDO          '5'

#define T_QUIT		'6'		/* General messages */
#define T_MAIL		'7'

#define MESSAGE_LENGTH	125

static char message_buffer[MESSAGE_LENGTH];
static char message_length[4];

static void receive_select (), receive_rotate (), receive_srand (),
	    receive_fire (), receive_next_turn(), receive_quit (),
	    receive_undo();

void end_communications()
{
	if (players[1].fd > 0) {
		XtRemoveInput (EnemyID);
		players[1].fd = -1;
	}
}

static void read_data (int fd, char *output, int length)
{
	extern int errno;

	if (read (fd, output, length) == length)
		return;

	if (errno != 0)
		perror ("read_data()");

	printf("Communications breakdown. Host/Program crash on other side?\n");

	XtDestroyApplicationContext (app_con);
	exit (1);
}

void EnemyInterrupt (XtPointer client_data,
				    int *source, XtInputId *id)
{
	if (players[1].fd == -1)
		return;

	read_data (players[1].fd, message_length, sizeof(message_length));
	read_data (players[1].fd, message_buffer, atoi (message_length));

	switch (message_buffer[0]) {
	case T_SELECT:
		receive_select ();
		break;
	case T_ROTATE:
                receive_rotate ();
		break;
	case T_FIRE:
		receive_fire ();
		break;
	case T_NEXT_TURN:
                receive_next_turn ();
		break;
	case T_SRAND:
		receive_srand ();
		break;
	case T_UNDO:
		receive_undo();
		break;
	case T_QUIT:
		end_communications ();
		receive_quit ();
		break;
	case T_MAIL:
		message ("%s: \"%s\"", players[1].name,
			 &message_buffer[2]);
		break;
	default:
		message ("Unknown input from opponent (type [%o] %c)",
			  message_buffer[0], message_buffer[0]);
	}
}

static void receive_undo()
{
  ::board->undo(0);
}

static void receive_quit ()
{
  message ("You win %s, because %s admitted your superior strategy!", 
	   players[0].name, players[1].name);  
  quit ();
}

static void receive_select ()
{
  int y, x;
  
  sscanf (&message_buffer[1], "%d %d", &x, &y);
  ::board->select_cb(0,x,y);
}

static void receive_rotate ()
{
  ::board->rotate_cb(0);
}

static void receive_fire ()
{
  ::board->fire_laser(0);
}

static void receive_next_turn ()
{
  ::board->next_turn(0);
}

static void receive_srand ()
{
  int s;
  
  sscanf(&message_buffer[1], "%d", &s);
  goodsrandom(s);
}

/*
 * Send module.
 */

/*VARARGS0*/
static void send_buffer (char *format, ...)
{
	va_list arglist;

	va_start(arglist, format);

	vsprintf (message_buffer, format, arglist);
	va_end(arglist);

	sprintf(message_length, "%d", strlen(message_buffer) + 1);

	write(players[1].fd, message_length, sizeof (message_length));
	write(players[1].fd, message_buffer, atoi (message_length));
}

void send_select (int x, int y)
{
  send_buffer ("%c %d %d", T_SELECT, x, y);
}

void send_rotate ()
{
  send_buffer ("%c", T_ROTATE);
}

void send_fire ()
{
  send_buffer ("%c", T_FIRE);
}

void send_next_turn ()
{
  send_buffer ("%c", T_NEXT_TURN);
}

void send_srand(int s)
{
  send_buffer("%c %d", T_SRAND, s);
}

void send_quit()
{
  send_buffer("%c", T_QUIT);
}

void send_undo()
{
  send_buffer("%c",T_UNDO);
}






