#include "alc.h"
#include "alcboard.h"
#include "pieces.h"

struct AlcPiece pieces[]={
  {"King",          1, 1, 0, 55, 56},
  {"Block",         8, 1, 0, 20, 28},
  {"Mirror Block",  1, 1, 0, 38, 19},
  {"Laser",         8, 0, 0, 77, 85},
  {"Freezer",       8, 0, 0,  1,  9},
  {"Bomb",          2, 0, 0, 17, 36},
  {"Shield",        8, 0, 0, 39, 47},
  {"Mirror",        8, 0, 0, 58, 66},
  {"Beam Splitter", 8, 0, 0, 96,104},
  {"Teleporter",    1, 0, 1, 74, 75},
  {"Hyperhole",     1, 0, 0, 57, 57}
};
#define Npieces (sizeof(pieces)/sizeof(AlcPiece))

enum {
  PIECE_KING=0,PIECE_BLOCK,PIECE_MBLOCK,PIECE_LASER,PIECE_FREEZER,
  PIECE_BOMB,PIECE_SHIELD,PIECE_MIRROR,PIECE_BSPLITTER,PIECE_TELEPORTER,
  PIECE_HYPER
};

static int holes[5][2]={ {7,1}, {7,3}, {7,5}, {7,7}, {7,9} };
#define Nholes (sizeof(holes)/(2*sizeof(int)))


AlcSquare::AlcSquare() { init(); }
inline void AlcSquare::init() { set("0000"); hole=0; }
inline int AlcSquare::set(String id)
{
  if(id[0]=='0') { piece=color=rot=frozen=used=mark=0; return 1; }
  piece = id[0]-'a'; if(piece>=Npieces) goto FAIL;
  color = id[1]-'0'; if(color>3) goto FAIL;
  else { if(color==3) { color==0; hole=1; }
  else hole=0; }
  rot   = id[2]-'0'; if(rot>=pieces[piece].r) goto FAIL;
  if(id[3]<'0') goto FAIL;
  frozen= (id[3]-'0') & 077;
  used  = (id[3]-'0') & 0100;
  mark = 0;
  return 1;
 FAIL:
  XtWarningMsg(id,"bad square initialization string",
	       XtCString,"AlcSquare::set():BadString",
	       NULL, 0);
  return 0;
}
  
String AlcSquare::get(char *id)
{
  id[0]=empty()?'0':piece+'a';
  id[1]=(hole?3:color)+'0';
  id[2]=rot+'0';
  id[3]=((frozen&077) | used?0100:0)+'0';
  id[4]=0;
  return id;
}

inline void AlcSquare::operator = (AlcSquare &s)
{
  piece = s.piece;
  color = s.color;
  rot = s.rot;
  frozen = s.frozen;
  used = s.used;
}

inline int AlcSquare::icon()
{
  switch(color){
  case 0: if(piece==PIECE_HYPER)
    return 0x200+(hole?0:pieces[piece].c1);
  else return hole?0x200:0;
	case 1: return (pieces[piece].c1+rot)+(frozen?0x100:0)+(hole?0x200:0);
	case 2: return (pieces[piece].c2+rot)+(frozen?0x100:0)+(hole?0x200:0);
	default: return -1;
	}
}

inline void AlcSquare::freeze()
{
  frozen+=1+goodrandom(7);
}

inline void AlcSquare::rotate()
{
  rot=(rot+1)%pieces[piece].r;
}

inline int AlcSquare::empty() { 
  return hole || ( color==0 && piece!=PIECE_HYPER);
};

inline int AlcSquare::eater() { return pieces[piece].eater; }
inline int AlcSquare::teleporter() { return pieces[piece].teleporter; }
inline int AlcSquare::rotatable() { return pieces[piece].r>1; }

static void select_cb_stub(Widget w,XtPointer object,XtPointer data)
{
  ((AlcBoard *)object)->
    select_cb(1,
	      ((AlcBoardCallbackStruct *)data)->x,
              ((AlcBoardCallbackStruct *)data)->y); 
}
static void rotate_cb_stub(Widget w,XtPointer object,XtPointer data)
{ ((AlcBoard *)object)->rotate_cb(1); }

AlcBoard::AlcBoard(Widget w,void (*showmovefnc)(AlcBoard &))
{
  boardw = w;
  XtVaGetValues(boardw, XtNdarkBackground,  &bg0,
		        XtNlightBackground, &bg1, NULL);
  init();
  XtAddCallback(boardw, XtNselectCallback, select_cb_stub, this);
  XtAddCallback(boardw, XtNadjustCallback, rotate_cb_stub, this);
  showmoves = showmovefnc;
  setup = 0;
}

void AlcBoard::init()
{
  int i,j;
  for(i=0;i<COLS;i++) for(j=0;j<ROWS;j++) square[i][j].init();
  selx=sely=-1;
  player=0;
  moves=3;
  lastlaser=0;
  rotateflag=0;
  undotype=-1;
}

void AlcBoard::sync()
{
  int i,j;
  for(i=0;i<COLS;i++) for(j=0;j<ROWS;j++)
    AlcBoardSetSquare(boardw,i,j,square[i][j].icon());
}

AlcBoard::load(char *filename)
{
  FILE *fp;
  int i,j,n;
  char str[100];

  deselect();
  if(!(fp=fopen(filename,"r"))){
    XtWarningMsg(filename, "cannot open file",
		 XtCFile, "AlcBoard::load():OpenFailed",
		 NULL,0);
    return 0;
  }

  undotype=-1;
  n=fscanf(fp,"XLC1\n%d %d",&moves,&player);
  if(n!=2) goto BADFILE;
  for(j=0;j<ROWS;j++) for(i=0;i<COLS;i++){
    n=fscanf(fp,"%s",str);
    if(n!=1) goto BADFILE;
    square[i][j].init();
    if(!square[i][j].set(str)) goto BADFILE;
  }
  fclose(fp);
  XtVaSetValues(boardw, XtNdarkBackground,  bg0,
		        XtNlightBackground, bg1, NULL);
  change_moves(0);
  refresh();
  return 1;
 BADFILE:
  fclose(fp);
  XtWarningMsg(filename, "bad file",
	       XtCFile, "AlcBoard::load():BadFile",
	       NULL,0);
  return 0;
}

AlcBoard::save(char *filename)
{
  FILE *fp;
  int i,j,n;
  char str[100];

  if(!(fp=fopen(filename,"w"))){
    XtWarningMsg(filename, "cannot create file",
		 XtCFile, "AlcBoard::save():OpenFailed",
		 NULL,0);
    return 0;
  }

  fprintf(fp,"XLC1\n%d %d\n",moves,player);
  for(j=0;j<ROWS;j++) {
    for(i=0;i<COLS;i++){
      square[i][j].get(str);
      fprintf(fp,"%s ",str);
    }
    fprintf(fp,"\n");
  }
  fclose(fp);
  return 1;
 BADFILE:
  fclose(fp);
  XtWarningMsg(filename, "save error",
	       XtCFile, "AlcBoard::save():SaveError",
	       NULL,0);
  return 0;
}

void AlcBoard::checkwinner()
{
  int king1=0;
  int king2=0;
  int i,j;
  Pixel c1,c2;

  for(j=0;j<ROWS;j++) for(i=0;i<COLS;i++) {
    king1 |= (square[i][j].piece==PIECE_KING && square[i][j].color==1);
    king2 |= (square[i][j].piece==PIECE_KING && square[i][j].color==2);
  }

  if(king1 && king2) return;
  if(king1) { c1=appData.player1Color; c2=appData.player1Color; }
  if(king2) { c1=appData.player2Color; c2=appData.player2Color; }
  if(!(king1 || king2))
    { c1=appData.player1Color; c2=appData.player2Color; }

  XtVaSetValues(boardw, XtNdarkBackground,  c1,
		        XtNlightBackground, c2, NULL);
  say(SAY_WIN);
  player=2;
}

void AlcBoard::change_moves(int diff)
{
  int i,x,y,n=0;
  int p[Nholes][2];

  for(i=0;i<Nholes;i++) {
    x=holes[i][0];y=holes[i][1];
    if(square[x][y].hole && square[x][y].color) {
      p[n][0]=x;p[n][1]=y;n++;
      square[x][y].set("0000");
      if(i==2) square[x][y].piece=PIECE_HYPER;
      AlcBoardSetSquare(boardw,x,y,square[x][y].icon());
      AlcBoardRefreshSquare(boardw,x,y,0);
    }
  }
  if(n > 0) { undotype=-1; say(SAY_EXPLODE); AlcBoardFireworks(boardw,n,p); }
  if(!setup) {
    moves += diff;
    checkwinner();
  }
  if(showmoves) showmoves(*this);
}

void AlcBoard::next_turn(int local)
{
  int i,j,x,y;
  
  deselect();
  if(player==2) { say(SAY_CANT); return; } // nothing if game is over

  for(j=0;j<ROWS;j++) for(i=0;i<COLS;i++) {
    // melt each frozen piece
    if(square[i][j].frozen && !(--square[i][j].frozen)) {
      AlcBoardSetSquare(boardw,i,j,square[i][j].icon());
      AlcBoardRefreshSquare(boardw,i,j,0);
    }
    // and set it to unused
    square[i][j].used=0;
  }
  // open and close holes
  for(i=0;i<Nholes;i++){
    x=holes[i][0];y=holes[i][1];
    if(goodrandom(2)) {
      square[x][y].hole=!square[x][y].hole;
      AlcBoardSetSquare(boardw,x,y,square[x][y].icon());
      AlcBoardRefreshSquare(boardw,x,y,0);
    }
  }
  if(local && comm) send_next_turn();
  moves=3;
  player=1-player;
  change_moves(0);
  undotype=-1;
}

void AlcBoard::deselect()
{
  if(selx==-1) return;
  AlcBoardSelectSquare(boardw,0,0,0); // deselect
  selx=sely=-1;
  rotateflag=0;
  if(moves==0) next_turn(0);
}


void AlcBoard::select(int x,int y)
{
  AlcBoardSelectSquare(boardw,x,y,1);
  selx=x;sely=y;
}

void AlcBoard::undo(int local)
{
  int t;
  switch(undotype){
  case 0: // move
    move0(u0,u1,ux,uy);
    ud=-ud;
    change_moves(ud);
    break;
  case 1:
    t=square[ux][uy].rot;
    rotate0(ux,uy,u0);
    u0=t;
    ud=-ud;
    change_moves(ud);
    break;
  default:
    say(SAY_CANT);
    return;
  }
  if(local && comm) send_undo();
  deselect();
}    

void AlcBoard::select_cb(int local,int x,int y)
{
  if(selx!=-1){
    if(selx==x && sely==y) deselect();
    else if(allowmove(selx,sely,x,y)) {
      if(move(selx,sely,x,y)) {
	deselect();
      }
    } else say(SAY_CANT);
  }
  else if(allowselect(x,y)) {
    select(x,y);
  }
  else say(SAY_CANT);
  if(local && comm) send_select(x,y);
}

void AlcBoard::rotate_cb(int local)
{
  if(selx!=-1 && square[selx][sely].rotatable()) {
    if(!rotateflag) { undotype=1; change_moves(-1); 
		      ud=-1; ux=selx; uy=sely; u0=square[selx][sely].rot; }
    square[selx][sely].rotate();
    rotate0(selx,sely,-1);
    rotateflag++;
    if(rotateflag == pieces[square[selx][sely].piece].r) {
      rotateflag=0; undotype=-1; change_moves(1);
    }
    if(local && comm) send_rotate();
  } else say(SAY_CANT);
}

void AlcBoard::rotate0(int x, int y, int r)
{
  if(r>=0) square[x][y].rot=r;
  AlcBoardSetSquare(boardw,x,y,square[x][y].icon());
  AlcBoardRefreshSquare(boardw,x,y,0);
}

void AlcBoard::move0(int x0, int y0, int x1, int y1)
{
    square[x1][y1]=square[x0][y0];
    square[x0][y0].set("0000");
    rotate0(x0,y0,-1);
    rotate0(x1,y1,-1);
    ux=x0; uy=y0; u0=x1; u1=y1;
}
  
int AlcBoard::move(int x0, int y0, int x1, int y1)
{
  undotype=-1;
  if(square[x1][y1].piece==PIECE_HYPER && !square[x1][y1].hole) {
    if(setup || !square[x1][y1].used) {
      teleport(x0,y0); square[x1][y1].used=1;
    }
    else { say(SAY_CANT); return 0; }
  } else {
    if(square[x0][y0].teleporter() && !square[x1][y1].empty()) {
      if(setup || !square[x0][y0].used) { teleport(x1,y1); square[x0][y0].used=1; }
      else { say(SAY_CANT); return 0; }
    } else if(square[x0][y0].eater() && !square[x1][y1].empty()) {
      if(setup || !square[x0][y0].used) { say(SAY_CRUNCH); square[x0][y0].used=1; }
      else { say(SAY_CANT); return 0; }
    } else undotype=0;
    move0(x0,y0,x1,y1);
  }

  ud=-(abs(x1-x0)+abs(y1-y0));
  change_moves(ud);
  return 1;
}

void AlcBoard::teleport(int x0,int y0)
{
  int x,y,n;
  for(n=y=0;y<ROWS;y++) for(x=0;x<COLS;x++) if(square[x][y].empty()) n++;
  n=goodrandom(n);
  for(y=0;y<ROWS;y++) for(x=0;x<COLS;x++) {
    if(square[x][y].empty()) n--;
    if(n==0) goto BREAK;
  }
  
 BREAK:
  say(SAY_HYPERIN);
  move0(x0,y0,x,y);
  rotate0(x,y,goodrandom(pieces[square[x][y].piece].r));
  say(SAY_HYPEROUT);
}

int  AlcBoard::allowmove(int x0, int y0, int x1, int y1)
{
  int c,p,h,r;
  if(setup) return 1;
  if(square[x1][y1].empty() ||
     square[x1][y1].piece == PIECE_HYPER ||
     square[x0][y0].eater() ||
     square[x0][y0].teleporter() ) {
    c=square[x1][y1].color; square[x1][y1].color=0;
    p=square[x1][y1].piece; square[x1][y1].piece=0;
    h=square[x1][y1].hole; square[x1][y1].hole=0;
    r=moveable(x0,y0,x1,y1,moves);
    square[x1][y1].color=c;
    square[x1][y1].piece=p;
    square[x1][y1].hole=h;
    return r;
  }
  else return 0;
}

int AlcBoard::moveable(int x0,int y0,int x1,int y1,int moves)
{
  if(setup) return 1;
  if(moves<0) return 0;
  if(x0==x1 && y0==y1) return 1;
  return
    (x0<x1 && square[x0+1][y0].empty() && !square[x0+1][y0].hole && 
     moveable(x0+1,y0,x1,y1,moves-1)) ||
    (x0>x1 && square[x0-1][y0].empty() && !square[x0-1][y0].hole && 
     moveable(x0-1,y0,x1,y1,moves-1)) ||
    (y0<y1 && square[x0][y0+1].empty() && !square[x0][y0+1].hole &&
     moveable(x0,y0+1,x1,y1,moves-1)) ||
    (y0>y1 && square[x0][y0-1].empty() && !square[x0][y0-1].hole && 
     moveable(x0,y0-1,x1,y1,moves-1)) ;
}

int AlcBoard::allowselect(int x, int y)
{
  if(square[x][y].color==0) return 0;
  if(setup) return 1;
  if(square[x][y].color-1 != player) return 0;
  return !square[x][y].frozen;
}

inline int AlcBoard::exitlaser(int x, int y, int direction)
{
  int r = square[x][y].mark & (1<<direction);
  square[x][y].mark |= (1<<direction);
  if(!r) AlcBoardDrawLaserSegments(boardw,x,y,1<<direction);
  return !r;
}

void AlcBoard::enterlaser(int x, int y, int direction)
{
  square[x][y].mark |= (1L<<8);
  AlcBoardDrawLaserSegments(boardw,x,y,1<<((direction+4)&7));
}

int AlcBoard::hit(int x, int y, int direction)
{
  int d=((direction+8-square[x][y].rot)&7);

#define backfire fire_laser(x,y,direction)
#define passthru fire_laser(x,y,(direction+4)&7)
#define frozen (square[x][y].frozen)
  switch(square[x][y].piece){
  case PIECE_KING:
  case PIECE_LASER:
    return 0;
  case PIECE_BLOCK:
    if( (d==7 || d==0 || d==1) && !frozen) { backfire; return 1; }
    return 0;
  case PIECE_MBLOCK:
    if(!frozen) { backfire; return 1; }
    return 0;
  case PIECE_FREEZER:
    if( (d==6 || d==0 || d==2) && !frozen) { backfire; return 1; }
    return 0;
  case PIECE_BOMB:
#define bombhit(a,b,dir) \
    if(!(square[a][b].empty() || square[a][b].piece==PIECE_HYPER) ) {\
	exitlaser(x,y,dir); enterlaser(a,b,dir);\
	square[a][b].mark |= (1L<<12);\
	AlcBoardSetSquare(boardw,a,b,square[a][b].icon()|0x100);\
	AlcBoardRefreshSquare(boardw,a,b,1);\
      }
    if(frozen || (d&1)) return 0;
    if(x>0) {
      if(y>0) bombhit(x-1,y-1,7);
      if(y<(ROWS-1)) bombhit(x-1,y+1,5);
      bombhit(x-1,y,6);
    }
    if(x<(COLS-1)) {
      if(y>0) bombhit(x+1,y-1,1);
      if(y<(ROWS-1)) bombhit(x+1,y+1,3);
      bombhit(x+1,y,2);
    }
    if(y>0) bombhit(x,y-1,0);
    if(y<(ROWS-1)) bombhit(x,y+1,4);
    return 0;
#undef bombhit
  case PIECE_SHIELD:
    if(frozen) return 0;
    if(d==0) { backfire; return 1; }
    if(d==4) { passthru; return 1; }
    return 0;
  case PIECE_MIRROR:
    if(d==6 || d==2) { passthru; return 2; }
    if(frozen) return 0;
    if(d==0) { backfire; return 1; }
    if(d==7) { fire_laser(x,y,(direction+2)&7); return 1; }
    if(d==1) { fire_laser(x,y,(direction+6)&7); return 1; }
    return 0;
  case PIECE_BSPLITTER:
    if(d&1) { passthru; return 2; }
    if(frozen || d==4) return 0;
    if(d==0 || d==6) fire_laser(x,y,(direction+2)&7);
    if(d==0 || d==2) fire_laser(x,y,(direction+6)&7);
    return 2;
  case PIECE_TELEPORTER:
    if(frozen) return 0;
    fire_laser(x,y,goodrandom(8));
    return 2;
  case PIECE_HYPER:
    if(square[x][y].hole) { passthru; return 1; }
    fire_laser(x,y,goodrandom(8));
    return 2;
  }

#undef frozen
#undef passthru
#undef backfire
  return 2;
}

void AlcBoard::fire_laser(int x,int y,int direction)
{
  do {
    if(!exitlaser(x,y,direction)) return;
    switch(direction){
      case 0: y--; break;
      case 1: x++;y--; break;
      case 2: x++; break;
      case 3: x++;y++; break;
      case 4: y++; break;
      case 5: x--;y++; break;
      case 6: x--; break;
      case 7: x--;y--; break;  
    }
    if(x<0 || x>=COLS || y<0 || y>=ROWS) return;
    enterlaser(x,y,direction);
  } while(square[x][y].empty());
  switch(hit(x,y,(direction+4)&7)){
  case 0: /* figure was destroyed */
    square[x][y].mark |= (1L<<12);
    AlcBoardSetSquare(boardw,x,y,square[x][y].icon()|0x100);
    AlcBoardRefreshSquare(boardw,x,y,1);
    break;
  case 1: /* figure stays and is redrawn */
    AlcBoardRefreshSquare(boardw,x,y,1);
    break;
  case 2: /* figure stays */
    break;
  }
}

void AlcBoard::fire_laser(int local)
{
  switch(lastlaser) { 
  case 1:
  case 2:
    if(local && comm) send_fire();
    clean_laser(); 
    return;
  case 3:
    lastlaser=0;
    return;
  }
  if(selx == -1 || square[selx][sely].used || moves<=0) {
    say(SAY_CANT);
    lastlaser=3;
    return;
  }
  if(square[selx][sely].piece == PIECE_LASER) {
    lastlaser=1;
    square[selx][sely].used=1;
    fire_laser(selx,sely,square[selx][sely].rot);
    AlcBoardRefreshSquare(boardw,selx,sely,1);
    undotype=-1;
  } else if(square[selx][sely].piece == PIECE_FREEZER) {
    lastlaser=2;
    square[selx][sely].used=1;
    fire_laser(selx,sely,(square[selx][sely].rot+4)&7);
    AlcBoardRefreshSquare(boardw,selx,sely,1);
    undotype=-1;
  } else {
    say(SAY_CANT);
    lastlaser=3;
  }
  if(local && comm) send_fire();
}

void AlcBoard::clean_laser()
{
  int i,j,n=0;
  int debris[36][2];

  for(i=0;i<COLS;i++) for(j=0;j<ROWS;j++) {
    if(square[i][j].mark & (1L<<12)){
      switch(lastlaser){
      case 2: // freezer
	square[i][j].freeze();
	break;
      case 1: // real laser
	debris[n][0]=i; debris[n][1]=j; n++;
	square[i][j].set("0000");
	AlcBoardSetSquare(boardw,i,j,0);
	AlcBoardRefreshSquare(boardw,i,j,0);
	break;
      default:
	fprintf(stderr,"lastlaser=%d\n",lastlaser);
      }
    }
    if(square[i][j].mark) {
      AlcBoardRefreshSquare(boardw,i,j,0);
      square[i][j].mark = 0;
    }
  }
  if(n>0) { say(SAY_EXPLODE); AlcBoardFireworks(boardw,n,debris); }
  change_moves(-1);
  deselect();
  lastlaser=0;
}
