#include "alcboard.h"

struct AlcPiece {
  char *name;
  int  r;      /* number of rotations */
  int eater, teleporter;
  int  c1,c2;  /* icon numbers for red & green */
};

struct AlcSquare {
  unsigned int piece;
  unsigned int color;
  unsigned int rot;
  unsigned int frozen;
  unsigned int mark;
  unsigned int used;
  unsigned int hole;
  
  AlcSquare();
  inline void init();
  inline int set(String id);
  String get(char *);
  inline void operator = (AlcSquare &s);
  inline int icon();
  inline void freeze();
  inline void rotate();
  inline int empty();
  inline int eater();
  inline int teleporter();
  inline int rotatable();
};

struct AlcBoard { /* AlcGame, actually */
  AlcSquare square[15][11];
  Widget boardw;
  Pixel bg0,bg1;
  int selx,sely;
  int player;        /* player who has a turn */
  int thisplayer;    /* this player */
  int moves;
  int setup;         /* setup mode */
  int rotateflag;    /* whether currently selected piece has been rotated */
  int lastlaser;     /* last laser fired: 0 none, 1 real, 2 freezer */
  void (*showmoves)(AlcBoard &); /* function which shows how many moves left */
  int undotype;      /* 0 = move, 1 = rotate */
  int ux,uy,u0,u1,ud;/* undo variables */

  AlcBoard(Widget w,void (*showmovefnc)(AlcBoard &));
  void init();
  int load(char *);
  int save(char *);
  void checkwinner();
  void sync();
  void refresh() { sync(); AlcBoardRefresh(boardw); }

  void select_cb(int local, int x, int y);
  void rotate_cb(int local);
  void undo(int local);
  void move0(int x0, int y0, int x1, int y1);
  void rotate0(int x, int y, int r);

  void deselect();
  void select(int x,int y);
  int  move(int x0, int y0, int x1, int y1);
  void teleport(int x,int y);
  void fire_laser(int x,int y,int direction);
  void fire_laser(int local);
  void clean_laser();
  void next_turn(int local);
  void change_moves(int diff);

  inline void enterlaser(int x, int y, int direction);
  int exitlaser(int x, int y, int direction);
  int hit(int x, int y, int direction);

  int  allowmove(int x0, int y0, int x1, int y1);
  int  moveable(int x0,int y0,int x1,int y1,int moves);
  int  allowselect(int x,int y);
};
